@@ -1,4 +1,4 @@
-# $Id: XSV.pm 36 2006-10-03 03:29:55Z zackse $
+# $Id: XSV.pm 43 2006-10-13 15:45:07Z zackse $
 
 package CGI::Application::Plugin::Output::XSV;
 
@@ -25,101 +25,153 @@ our %EXPORT_TAGS= (
   all => [ @EXPORT, @EXPORT_OK ],
 );
 
-our $VERSION= '0.9';
+our $VERSION= '1.00';
 
 ##
 
 sub xsv_report {
-  my $args= shift || {};
+  my $args = shift || {};
 
   croak "argument to xsv_report must be a hash reference"
     if ref( $args ) ne 'HASH';
 
-  my %defaults= (
+  my %defaults = (
     headers         => undef,
     headers_cb      => \&clean_field_names,
     include_headers => 1,
     fields          => undef,
     values          => undef,
-    get_row_cb      => undef,
+    row_filter      => undef,
+    iterator        => undef,
     line_ending     => "\n",
     csv_opts        => {},
+    maximum_iters   => 1_000_000, # XXX reasonable default?
   );
 
-  my %opts= ( %defaults, %$args );
+  my %opts = ( %defaults, %$args );
 
-  croak "need array reference values to do anything"
-    unless $opts{values} && ref( $opts{values} ) eq 'ARRAY';
+  # deprecated option
+  if ( $opts{get_row_cb} ) {
+    if ( $opts{row_filter} ) {
+      carp "ignoring use of deprecated get_row_cb when row_filter specified";
+    }
+    else {
+      $opts{row_filter} = $opts{get_row_cb};
+      carp "get_row_cb is deprecated, please use row_filter instead";
+    }
+  }
+
+  croak "need array reference of values or iterator to do anything"
+    if   ! ( $opts{values}   && ref( $opts{values} )   eq 'ARRAY' )
+      && ! ( $opts{iterator} && ref( $opts{iterator} ) eq 'CODE'  );
+
+  croak "can't supply both values and iterator"
+    if $opts{values} && $opts{iterator};
 
   # list of fields to include in report
-  my $fields= [];
+  my $fields = [];
 
-  if( $opts{fields} ) {
-    $fields= $opts{fields};
+  if ( $opts{fields} ) {
+    # user-specified
+    $fields = $opts{fields};
   }
-  elsif( @{ $opts{values} } ) {
-    my $list_type= ref( $opts{values}[0] );
+  elsif ( $opts{values} ) {
+    # try to determine field names from provided values
+    if ( @{ $opts{values} } ) {
+      my $list_type = ref( $opts{values}[0] );
 
-    # field list from first entry in value list
-    if( $list_type eq 'HASH' ) {
-      $fields= [ keys %{ $opts{values}[0] } ];
+      # field list from first entry in value list
+      if ( $list_type eq 'HASH' ) {
+        $fields = [ keys %{ $opts{values}[0] } ];
+      }
+      # or simply array indices
+      elsif ( $list_type eq 'ARRAY' ) {
+        $fields = [ 0..$#{$opts{values}[0]} ];
+      }
+      else {
+        croak "unknown list type [$list_type]";
+      }
     }
-    # or simply array indices
-    elsif( $list_type eq 'ARRAY' ) {
-      $fields= [ 0..$#{$opts{values}[0]} ];
+    else {
+      croak "can't determine field names (values is an empty list), aborting";
+    }
+  }
+  else {
+    # using iterator, empty field list
+  }
+
+  # function to filter each row of data from $opts{values}
+  my $row_filter;
+
+  if ( $opts{row_filter} ) {
+    # user-specified
+    $row_filter = $opts{row_filter};
+  }
+  elsif ( $opts{values} && @{ $opts{values} } ) {
+    # simple defaults for slices
+    my $list_type = ref( $opts{values}[0] );
+
+    if ( $list_type eq 'HASH' ) {
+      $row_filter = sub { my ($row, $fields)= @_; return [ @$row{@$fields} ] };
+    }
+    elsif ( $list_type eq 'ARRAY' ) {
+      $row_filter = sub { my ($row, $fields)= @_; return [ @$row[@$fields] ] };
     }
     else {
       croak "unknown list type [$list_type]";
     }
   }
   else {
-    croak "can't determine field names (values is an empty list), aborting";
+    # using iterator, no filter
+    $row_filter = sub { $_[0] };
   }
 
-  # function to retrieve each row of data from $opts{values}
-  my $get_row;
-
-  if( $opts{get_row_cb} ) {
-    $get_row= $opts{get_row_cb};
-  }
-  elsif ( @{ $opts{values} } ) {
-    my $list_type= ref( $opts{values}[0] );
-
-    if( $list_type eq 'HASH' ) {
-      $get_row= sub { my( $row, $fields )= @_; return [ @$row{@$fields} ] };
-    }
-    elsif( $list_type eq 'ARRAY' ) {
-      $get_row= sub { my( $row, $fields )= @_; return [ @$row[@$fields] ] };
-    }
-    else {
-      croak "unknown list type [$list_type]";
-    }
-  }
-  else {
-    # empty values list -- always return empty list
-    $get_row= sub { return [] };
-  }
-
-  my $csv= Text::CSV_XS->new( $opts{csv_opts} );
+  my $csv = Text::CSV_XS->new( $opts{csv_opts} );
   my $output;
 
-  if( $opts{include_headers} ) {
-    if( ! $opts{headers} &&
-        !($opts{headers_cb} && ref( $opts{headers_cb} ) eq 'CODE')
-    )
-    {
-      croak "need headers or headers_cb to include headers";
+  if ( $opts{include_headers} ) {
+    if ( ! $opts{headers} ) {
+      if ( ! ($opts{headers_cb} && ref( $opts{headers_cb} ) eq 'CODE') ) {
+        croak "need headers or headers_cb to include headers";
+      }
+      elsif ( ! @{$fields} ) {
+        carp "passing empty fields list to headers_cb";
+      }
     }
 
     # formatted column headers
-    my $readable_headers= $opts{headers} || $opts{headers_cb}->( $fields )
+    my $readable_headers = $opts{headers} || $opts{headers_cb}->( $fields )
       or croak "can't generate headers";
 
+    croak "return value from headers_cb is not an array reference, aborting"
+      if ref ( $readable_headers ) ne 'ARRAY';
+
     $output .= add_to_xsv( $csv, $readable_headers, $opts{line_ending} );
   }
 
-  foreach( @{ $opts{values} } ) {
-    $output .= add_to_xsv( $csv, $get_row->($_, $fields), $opts{line_ending} );
+  if ( $opts{values} ) {
+    foreach my $list_ref ( @{ $opts{values} } ) {
+      $output .= add_to_xsv(
+        $csv, $row_filter->($list_ref, $fields), $opts{line_ending}
+      );
+    }
+  }
+  # using iterator
+  else {
+    my $iterations = 0;
+
+    while ( my $list_ref = $opts{iterator}->($fields) ) {
+      croak "return value from iterator is not an array reference, aborting"
+        if ref( $list_ref ) ne 'ARRAY';
+
+      # XXX infinite loop?
+      croak "iterator exceeded maximum iterations ($opts{maximum_iters})"
+        if ++$iterations > $opts{maximum_iters};
+
+      $output .= add_to_xsv(
+        $csv, $row_filter->($list_ref, $fields), $opts{line_ending}
+      );
+    }
   }
 
   return $output;
@@ -129,17 +181,17 @@ sub xsv_report {
 # same params as xsv_report, plus
 #   filename => 'download.csv',
 sub xsv_report_web {
-  my( $self, $args )= @_;
+  my ($self, $args) = @_;
   $args ||= {};
 
   croak "argument to xsv_report_web must be a hash reference"
     if ref( $args ) ne 'HASH';
 
-  my %defaults= (
+  my %defaults = (
     filename => 'download.csv',
   );
 
-  my %opts= ( %defaults, %$args );
+  my %opts = ( %defaults, %$args );
 
   $self->header_props(
     -type                  => 'application/x-csv',
@@ -150,11 +202,13 @@ sub xsv_report_web {
   return xsv_report( \%opts );
 }
 
+# default field name generator:
+#   underscores to spaces, upper case first letter of each word
 sub clean_field_names {
-  my $fields= shift;
+  my $fields = shift;
 
   # using temp var to avoid modifying $fields
-  my @fields_copy= @$fields;
+  my @fields_copy = @{$fields};
 
   return [
     map { tr/_/ /; s/\b(\w+)/\u$1/g; $_ } @fields_copy
@@ -162,17 +216,17 @@ sub clean_field_names {
 }
 
 sub add_to_xsv {
-  my( $csv, $fields, $line_ending )= @_;
-  croak "fields argument (required) must be an array reference"
-    unless $fields && ref( $fields ) eq 'ARRAY';
+  my ($csv, $fields, $line_ending) = @_;
+  croak "add_to_xsv: fields argument (required) must be an array reference"
+    if ! ($fields && ref( $fields ) eq 'ARRAY');
 
   # XXX redundant for empty string (or 0)
   $line_ending ||= '';
 
-  return $line_ending unless @$fields;
+  return $line_ending if ! @{$fields};
 
-  $csv->combine( @$fields )
-    or croak "Failed to add [@$fields] to csv: " . $csv->error_input();
+  $csv->combine( @{$fields} )
+    or croak "Failed to add [@{$fields}] to csv: " . $csv->error_input();
 
   return $csv->string() . $line_ending;
 }
@@ -193,9 +247,9 @@ CGI::Application::Plugin::Output::XSV - generate csv output from a CGI::Applicat
   # in some runmode...
 
   # $sth is a prepared DBI statement handle
-  my $members= $sth->fetchall_arrayref( {} );
+  my $members = $sth->fetchall_arrayref( {} );
 
-  my @headers= qw( member_id first_name last_name ... );
+  my @headers = qw( member_id first_name last_name ... );
 
   return $self->xsv_report_web({
     fields     => \@headers,
@@ -204,6 +258,17 @@ CGI::Application::Plugin::Output::XSV - generate csv output from a CGI::Applicat
     filename   => 'members.csv',
   });
 
+
+  # or, generate the list on the fly:
+
+  sub get_members { $sth->fetchrow_arrayref() }
+
+  return $self->xsv_report_web({
+    iterator   => \&get_members,
+    csv_opts   => { sep_char => "\t" },
+    filename   => 'members.csv',
+  });
+
 =head1 DESCRIPTION
 
 C<CGI::Application::Plugin::Output::XSV> provides csv-related routines
@@ -221,27 +286,27 @@ You may export all four routines by specifying the export tag C<:all>:
 
 On many websites, I had code to retrieve a list of data items for use
 in an L<HTML::Template|HTML::Template(3)> TMPL_LOOP. Usually this code
-would use the L<DBI|DBI(3)> routine C<fetchall_arrayref( {} )> to get a
+would use the L<DBI|DBI(3)> routine C<fetchall_arrayref()> to get a
 list of hash references, one for each data item.
 
-  my $users= $sth->fetchall_arrayref( {} );
+  my $users = $sth->fetchall_arrayref( {} );
 
-  my $template= $self->load_tmpl( ... );
+  my $template = $self->load_tmpl( ... );
 
   $template->param( users => $users );
 
   return $template->output;
 
-Inevitably, the client would ask for a data format they could load
-in Excel, so I'd add another runmode for a csv export. This runmode
-almost always looked the same:
+At some point, it would become apparent that an "export to spreadsheet"
+feature would be useful, so I'd add another runmode, which almost always
+looked the same:
 
-    my @fields= qw(keys to each data item);
+    my @fields = qw(keys to each data item);
 
-    my $csv= Text::CSV_XS->new();
+    my $csv = Text::CSV_XS->new();
 
-    foreach( @$users ) {
-      $csv->combine( [ @$_{ @fields } ] );
+    foreach my $user ( @{$users} ) {
+      $csv->combine( [ @{$user}{@fields} ] );
       $output .= $csv->string() . "\n";
     }
 
@@ -257,7 +322,7 @@ that wraps the above code while offering a fair amount of programmer
 flexibility.
 
 For example, the programmer may control the naming of header columns,
-filter each line of output before it is passed to L<Text::CSV_XS|Text::CSV_XS(3)>,
+filter each row of data before it is passed to L<Text::CSV_XS|Text::CSV_XS(3)>,
 and set the filename that is supplied to the user's browser.
 
 Please see the documentation below for C<xsv_report_web> for a list of
@@ -269,18 +334,20 @@ available options.
 
 =item B<xsv_report_web>
 
+  ## METHOD 1. Pre-generated list of values for csv
+
   # in a runmode
 
-  my @members= (
+  my @members = (
     { member_id  => 1,
       first_name => 'Chuck',
       last_name  => 'Berry', },
     ...
   );
 
-  my @headers= ("Member ID", "First Name", "Last Name");
+  my @headers = ("Member ID", "First Name", "Last Name");
 
-  my @fields = qw(member_id first_name last_name);
+  my @fields  = qw(member_id first_name last_name);
 
   return $self->xsv_report_web({
     fields     => \@fields,
@@ -290,6 +357,22 @@ available options.
     filename   => 'members.csv',
   });
 
+
+  ## METHOD 2. Generate list on the fly
+
+  # in a runmode
+
+  sub get_members { $sth->fetchrow_arrayref() }
+
+  my @headers = ("Member ID", "First Name", "Last Name");
+
+  return $self->xsv_report_web({
+    headers    => \@headers,
+    iterator   => \&get_members,
+    csv_opts   => { sep_char => "\t" },
+    filename   => 'members.csv',
+  });
+
 This method generates a csv file that is sent directly to the user's
 web browser. It sets the content-type header to 'application/x-csv' and sets
 the content-disposition header to 'attachment'.
@@ -298,7 +381,7 @@ It should be invoked through a
 L<CGI::Application|CGI::Application(3)> subclass object.
 
 It takes a reference to a hash of named parameters. All except for
-C<values> are optional:
+C<values> or C<iterator> are optional:
 
 =over 8
 
@@ -313,6 +396,9 @@ L<Text::CSV_XS|Text::CSV_XS(3)>. The default is an empty hash.
 
   fields => [ qw(member_id first_name last_name) ],
 
+  # or array indices
+  fields => [ 1, 2, 0 ],
+
 A reference to a list of field names or array indices. This parameter
 specifies the order of fields in each row of output.
 
@@ -323,6 +409,9 @@ order will be random because the field names are extracted from a hash.
 If the C<values> parameter is a list of lists, the field order will be
 the same as the data provided.
 
+If C<fields> is not supplied and C<iterator> is used instead of C<values>,
+the field list will be empty.
+
 =item filename
 
   filename => 'members.csv',
@@ -345,10 +434,10 @@ as a parameter to generate column headers.
 
   # replace underscores with spaces
   headers_cb => sub {
-    my $fields= shift;
+    my $fields = shift;
 
     # using temp var to avoid modifying $fields
-    my @fields_copy= @$fields;
+    my @fields_copy = @{$fields};
 
     return [
       map { tr/_/ /; $_ } @fields_copy
@@ -359,8 +448,8 @@ A reference to a subroutine used to generate column
 headers from the field names.
 
 A default routine is provided in C<clean_field_names>. This
-function is passed the list of fields (C<fields>) as a parameter
-and should return a reference to a list of column headers.
+function is passed a reference to the list of fields (C<fields>)
+as a parameter and should return a reference to a list of column headers.
 
 =item include_headers
 
@@ -391,23 +480,61 @@ The value appended to each line of csv output. The default is "\n".
   ],
 
 A reference to a list of hash references (such as
-that returned by the L<DBI|DBI(3)> C<fetchall_arrayref( {} )> routine, or
-a reference to a list of list references.
+that returned by the L<DBI|DBI(3)> C<fetchall_arrayref( {} )> routine),
+or a reference to a list of list references.
 
-This argument is required.
+Either this argument or C<iterator> must be provided.
 
-=item get_row_cb
+=item iterator
 
-  # uppercase all values -- assumes values are hash references
-  get_row_cb => sub {
-    my( $row, $fields )= @_;
+  iterator => sub { $sth->fetchrow_arrayref() },
 
-    return [ map { uc } @$row{@$fields} ];
+A reference to a subroutine that is used to generate each row
+of data. It is passed a reference to the list of fields (C<fields>)
+as a parameter and should return a reference to a list (which
+will be passed to C<add_to_xsv()>).
+
+It will be called repeatedly to generate each row of data until
+it returns a false value.
+
+This may be preferred to C<values> when the data set is large
+or expensive to generate up-front. Thanks to Mark Stosberg for
+suggesting this option.
+
+Either this argument or C<values> must be provided.
+
+=item maximum_iters
+
+  maximum_iters => 1_000_000,
+
+This is the maximum number of times the C<iterator> will be called
+before an exception is raised. This is a basic stopgap to
+prevent a runaway iterator that never returns false.
+
+The default is one million.
+
+=item row_filter
+
+  # uppercase all values
+  # (each row is a list of hash references)
+  row_filter => sub {
+    my ($row, $fields) = @_;
+
+    return [ map { uc } @{$row}{@$fields} ];
   },
 
-A reference to a subroutine used to generate each row of output
-(other than the header row). Default routines are provided that
-return each row of C<values> in the order specified by C<headers>.
+A reference to a subroutine used to filter each row of data
+(other than the header row) before it is passed to C<add_to_xsv>.
+When the C<values> parameter is
+supplied, a default filter is provided that produces each row
+in the order specified by C<headers>. For example, the default
+filter for a C<values> list of hash references is shown below.
+
+  row_filter => sub {
+    my ($row, $fields) = @_;
+
+    return [ @{$row}{@$fields} ];
+  },
 
 This subroutine is passed two parameters for each row:
 
@@ -423,6 +550,10 @@ the field list (C<fields> - reference to a list of hash keys or array indices)
 
 =back
 
+Note: This parameter used to be named C<get_row_cb>. That name is
+deprecated and a warning will be issued if it is used instead of
+C<row_filter>.
+
 =back
 
 =back
@@ -434,13 +565,13 @@ the field list (C<fields> - reference to a list of hash keys or array indices)
 =item B<add_to_xsv>
 
    # $sth is a prepared DBI statement handle
-   my $values= $sth->fetchall_arrayref( {} );
-   my @headers= qw/foo bar baz/;
+   my $values  = $sth->fetchall_arrayref( {} );
+   my @headers = qw/foo bar baz/;
    my $output;
 
    # $csv is a Text::CSV_XS object
-   foreach( @$values ) {
-      $output .= add_to_xsv( $csv, [ @$_{@headers} ], "\r\n" );
+   foreach my $href ( @{$values} ) {
+      $output .= add_to_xsv( $csv, [ @{$href}{@headers} ], "\r\n" );
    }
 
 This function, used internally by C<xsv_report>/C<xsv_report_web>,
@@ -470,13 +601,10 @@ On an error from L<Text::CSV_XS|Text::CSV_XS(3)>, the function raises an excepti
 On receiving an empty list of values, the function returns the
 line ending only.
 
-Should this return a formatted list of empty fields? Let me know if you
-think that would be better.
-
 =item B<clean_field_names>
 
-  my $fields= [ qw/first_name foo bar baz/ ];
-  my $headers= clean_field_names( $fields );
+  my $fields  = [ qw/first_name foo bar baz/ ];
+  my $headers = clean_field_names( $fields );
 
   # $headers is now [ 'First Name', 'Foo', 'Bar', 'Baz' ]
 
@@ -493,11 +621,11 @@ if the C<headers_cb> parameter is not supplied.
 =item B<xsv_report>
 
   # $sth is a prepared DBI statement handle
-  my $members= $sth->fetchall_arrayref( {} );
+  my $members = $sth->fetchall_arrayref( {} );
 
-  my @headers= qw( member_id first_name last_name ... );
+  my @headers = qw( member_id first_name last_name ... );
 
-  my $output= $self->xsv_report({
+  my $output = $self->xsv_report({
     fields     => \@headers,
     values     => $members,
     csv_opts   => { sep_char => "\t" },
@@ -593,7 +721,7 @@ which is not applicable to this function.
     ],
     fields          => [ qw(first_name last_name phone) ],
     headers_cb      => sub {
-      my @h= @{ +shift };
+      my @h = @{ +shift };
       s/_name$// foreach @h;
       return \@h;
     },
@@ -649,17 +777,17 @@ which is not applicable to this function.
 =item Filter data as it is processed
 
   sub plus_one {
-    my( $row, $fields )= @_;
+    my ($row, $fields) = @_;
 
-    return [ map { $_ + 1 } @$row{@$fields} ];
+    return [ map { $_ + 1 } @{$row}{@$fields} ];
   }
 
   # each row (other than header row) will be
   # passed through plus_one()
-  return  $self->xsv_report_web({
+  return $self->xsv_report_web({
     fields     => [ qw(foo bar baz) ],
     values     => [ { foo => 1, bar => 2, baz => 3 }, ],
-    get_row_cb => \&plus_one,
+    row_filter => \&plus_one,
   });
 
   __END__
@@ -683,18 +811,53 @@ which is not applicable to this function.
   1,2,3
   4,5,6
 
+=item Generate each row on the fly (streaming)
+
+  my @vals = qw(one two three four five six);
+
+  sub get_vals {
+    while ( @vals ) {
+      return [ splice @vals, 0, 3 ]
+    }
+  };
+
+  $self->xsv_report_web({
+    include_headers => 0,
+    iterator        => \&get_vals,
+  });
+
+  __END__
+  one,two,three
+  four,five,six
+
+=item Generate each row on the fly using a DBI iterator
+
+  my $get_vals = sub { $sth->fetchrow_arrayref() };
+
+  $self->xsv_report_web({
+    include_headers => 0,
+    iterator        => $get_vals,
+  });
+
+=item Use a DBI iterator, increment each value extracted
+
+  $self->xsv_report_web({
+    include_headers => 0,
+    iterator        => sub { $sth->fetchrow_arrayref() };
+    row_filter      => sub {
+      my $row = shift;
+      return [ map { $_ + 1 } @{$row} ];
+    },
+  });
+
 =back
 
 =head1 ERROR HANDLING
 
-=over 4
-
 The function C<add_to_xsv> will raise an exception when
 C<< Text::CSV_XS->combine >> fails. Please see the L<Text::CSV_XS|Text::CSV_XS(3)>
 documentation for details about what type of input causes a failure.
 
-=back
-
 =head1 AUTHOR
 
 Evan A. Zacks C<< <zackse@cpan.org> >>
@@ -713,13 +876,13 @@ L<Text::CSV_XS>, L<CGI::Application>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (c) 2005 CommonMind, LLC. All rights reserved.
+Copyright (c) 2006 CommonMind, LLC. All rights reserved.
 
 This program is free software; you can redistribute it
 and/or modify it under the same terms as Perl itself.
 
 =head1 REVISION
 
-$Id: XSV.pm 36 2006-10-03 03:29:55Z zackse $
+$Id: XSV.pm 43 2006-10-13 15:45:07Z zackse $
 
 =cut
